%% Example demonstrating different options for AMGCL
% Note that these choices mirror the names given in the AMGCL
% documentation: https://amgcl.readthedocs.io/en/latest/runtime.html
clear all
mrstModule add msrsb incomp coarsegrid spe10 linearsolvers

%Program that runs all combinations and prints results to a file

%% Create txt file
filename = '/home/shomea/m/mlnorhei/Desktop/MasterThesis/MRST_Sintef/mrst-2018a/mrst-2018a/MyTries/ExperimentalResults/olympus=1_relaxation'; %layers=1-5_AMG';
filenameError = strcat(filename,'_error.txt');
filename_mldivide = '/home/shomea/m/mlnorhei/Desktop/MasterThesis/MRST_Sintef/mrst-2018a/mrst-2018a/MyTries/ExperimentalResults/mldivide.txt';
filename = strcat(filename,'.txt');
fileID = fopen(filename,'w');
fileIDerror = fopen(filenameError,'w');
fileID_mldivide = fopen(filename_mldivide,'a');
% Without memory:
fprintf(fileID,'|%10s|%10s|%26s|%13s|%10s|%12s|%5s|%16s|\r\n','Precond','Solver','Coarsening','Relaxation','Iterations','Time','Conv?','Final residual');

results = zeros(7,4,8); % for 7 Krylov solvers, 4 coarsenings and 8 smoothers
hold on
%% Setup problem
simple = false;
hybrid = false;
olympus = true;

if simple
    model = 'Simple';
    % Create n by n grid
    n = 1000;
    h = 1;
    G = cartGrid([n, n, h]);
    G = computeGeometry(G);
    disp(G)
    rock = makeRock(G, 1*darcy, 1);
    W = [];
    W = addWell(W, G, rock, 1, 'type', 'rate', 'val', 1);
    %W = addWell(W, G, rock, 100, 'type', 'rate', 'val', 1);
    W = addWell(W, G, rock, G.cells.num, 'type', 'bhp', 'val', 0);
    
elseif hybrid
    model = 'Hybrid';
    % The simple case, but with the dimensions and well-setup of SPE10
    layerstart = 30;
    layerstop = 45;
    phi = 1; % porosity makes no difference for incompressible single phase flow
    layers = layerstart:layerstop;
    [G, W, ~] = getSPE10setup(layers);
    rock = makeRock(G,1*darcy,phi);
    
elseif olympus
    model = 'Olympus';
    mrstModule add deckformat
    caseNo = 1;
    casename = ['OLYMPUS_', num2str(caseNo)];
    fn = fullfile(mrstDataDirectory(), 'olympus', 'ECLIPSE', casename, [casename, '.DATA']);
    deck = readEclipseDeck(fn);
    deck = convertDeckUnits(deck);

    % Convert grid, rock and wells
    G = initEclipseGrid(deck);
    G = computeGeometry(G(1));

    rock  = initEclipseRock(deck);
    rock  = compressRock(rock, G.cells.indexMap);

    W = processWells(G, rock, deck.SCHEDULE.control(1));
    % Plot the wells + rock/perm
    figure;
    plotCellData(G, rock.poro);
    plotWell(G, W);

else
    model = 'SPE10';
    % Take part of SPE10 benchmark
    %layers = 1; % 1-35 smoother perm., 36-85 rougher and more varying
    layerstart = 21;
    layerstop = 50;
    layers = layerstart:layerstop;
    [G, W, rock] = getSPE10setup(layers);
end

% Set up linear system by calling the incompTPFA solver without a linear
% solver
fluid = initSingleFluid('rho', 0, 'mu', 1);
state = initResSol(G, 0);
state = incompTPFA(state, G, computeTrans(G, rock), fluid, 'Wells', W, 'matrixOutput', true, 'LinSolve', @(A, b) 0*b);
A = state.A;
b = state.rhs;

% Reduce to only wells
[A, b] = eliminateWellEquations(A, b, G.cells.num);

% Solve the system directly using mldivide, taking the time and writing to
% file:
tic();
%dummy = monitor_memory_whos;
%    x_dir = mldivide(A,b);
x_dir = A\b;
%mem_ml = monitor_memory_whos-dummy;
t_dir = toc();

conv = 'YES';
if (norm(b-A*x_dir) >= 1e-6) 
    conv = 'NO';
end
if olympus 
    layerstring = '';
    caseNo = num2str(caseNo);
    dim = '';
elseif simple
    layerstring = '';
    caseNo = '';
    dim = strcat(num2str(n), 'x', num2str(n), 'x', num2str(h));
else 
    if layerstart == layerstop
        layerstring = num2str(layerstart);
    else
        layerstring = strcat(num2str(layerstart),':',num2str(layerstop));
    end
    caseNo = '';
    dim = '';
end
% Without memory:
fprintf(fileID_mldivide,'|%10s|%10s|%10s|%15s|%13.2f|%5s|\r\n',model, caseNo, layerstring, dim, t_dir, conv);
fclose(fileID_mldivide);



% Make a choice between all alternative preconditioners, solvers,
% coarsenings and smoothers
makeChoice = @(text, choices)  listdlg('PromptString',  text,...
                                       'SelectionMode', 'single',...
                                       'ListString',    choices);
nt = 2; % the maximum number of computational threads





%spparms('spumoni',2) 








%% Select preconditioner
% There are two options here. If relaxation (AMG, you mean?) is chosen, AMG will be used as
% a preconditioner. If relaxation is chosen, no multigrid will be used and
% the solver will be a simpler Krylov solver with relaxation as a
% preconditioner. The coarsening option chosen later in this example will
% not be used in this case.
preconditioners = {'amg', ...
                   'relaxation'};
%index = makeChoice('Select solver', preconditioners);
for l=2:length(preconditioners)
    precond = preconditioners{l};
    %% Select solver
    % Choices for Krylov acceleration
    solvers = {'bicgstab', ... % Biconjugate gradient stabilized method
               'cg', ... % Conjugate gradient
               'bicgstabl', ... % Bicgstab variant
               'gmres', ... % Generalized minimal residual method
               'lgmres', ... % GMRES variant
               'fgmres', ... % Flexible-GMRES 
               'idrs'}; % Induced Dimension Reduction method IDR(s)
    %index = makeChoice('Select Krylov-solver', solvers);
    for k=1:length(solvers)
        solver = solvers{k};
        %% Select coarsening and interpolation scheme for AMG
        coarsening = {'ruge_stuben', ... % Ruge-stuben coarsening ("standard" coarsening)
                      'smoothed_aggregation', ... % Aggregation with smoothed iterates
                      'aggregation', ... % Aggregation - very fast coarsening due to very simple constant interpolation
                      'smoothed_aggr_emin'}; % Aggregation with energy-minimizing smoothing
        %index = makeChoice('Select coarsening', coarsening);
        for j=1:length(coarsening)-3
            coarsen = coarsening{j};
            %% Select relaxation ("smoother")
            relaxations = {'spai0', ... % Sparse approximate inverse, degree 0
                           'spai1', ... % Sparse approximate inverse
                           'gauss_seidel', ... % Gauss-seidel
                           'ilu0', ... % Incomplete LU-factorization without fill-in
                           'iluk', ... % Level-based ILU
                           'ilut', ... % Thresholded ILU
                           'damped_jacobi', ... % Damped Jacobi
                           'chebyshev'}; % Chebyshev smoothing
            %index = makeChoice('Select relaxation', relaxations);
            %relax = relaxations{index};

            %%%%%%%%
            for i = 1:length(relaxations)
                relax = relaxations{i};
                fprintf('============================================================================ \n %s, %s, %s, %s \n',precond,solver,coarsen,relax)
                %% Solve the system (but avoid bad combinations)
%      

                 try
                     tic();
                     %dummy = monitor_memory_whos;
                     maxNumCompThreads(nt)
                     maxIter = 5000;
                     [x, err, iter] = callAMGCL(A, b, ... 
                                        'coarsening', coarsen,...
                                        'relaxation', relax, ...
                                        'solver', solver, ...
                                        'maxIterations', maxIter, ...
                                        'preconditioner', precond, ...
                                        'verbose', true);
                     %mem = monitor_memory_whos-dummy;
                     t_amg = toc();
                     conv = 'YES';
                     results(k,j,i) = iter;
                     if (iter==5000) || (norm(b-A*x_dir) >= 1e-6 || isNaN(norm(b-A*x_dir))) 
                         conv = 'NO';
                     end
                     fprintf(fileID,'|%10s|%10s|%26s|%13s|%10i|%12.2f|%5s|%16d|\r\n',precond,solver,coarsen,relax,iter,t_amg,conv,err);
                 catch ME
                     message = ME.message;
                     fprintf(fileIDerror,'|%10s|%10s|%26s|%13s|\r\n%s\r\n',precond,solver,coarsen,relax,message);

                 end


            end
            fprintf(fileID,'|----------|----------|--------------------------|-------------|----------|------------|-----|----------------|------------------|\r\n');
        end
    end
end


%% Additional options
% There are many additional options which can be set through the AMGCL
% struct. Generally, these can be passed as keyword arguments to callAMGCL
% and other routines which rely on AMGCL.
% 
% Especially interesting options are
%  - ncycle (number of multigrid cycles)
%  - npre (number of presmootings)
%  - npost (number of postsmoothings)
% As well as options relating to the coarsening hierarchy:
%  - rs_eps_strong (strong threshold, ruge_stuben)
%  - aggr_eps_strong (strong threshold, aggregation)
%str = getAMGCLMexStruct();
%disp(str)